from argparse import RawTextHelpFormatter

from django.apps import apps
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.core.management.base import BaseCommand, CommandError
from django.db import connection, models, transaction
from mptt.fields import TreeForeignKey
from mptt.models import MPTTModel
from parler.models import TranslatableModel

from fluent_blogs import appsettings
from fluent_blogs.models import get_entry_model
from fluent_blogs.six import python_2_unicode_compatible


class Command(BaseCommand):
    """
    Migrate a the blog category model data and constraints.

    The ideal replacement is django-categories-i18n, which also allows the category names
    to be translated. This command migrates the data, and adjusts the foreign key in the M2M table.
    """
    args = "--from=APP.MODEL --to=APP.MODEL"
    help = (
        "Migrate categories to a new model. This can be used when the old project used django-categories,\n"
        "but you would like to migrate that to django-categories-i18n. Take the following steps:\n\n"
        " - make sure 'categories_i18n' is in INSTALLED_APPS.\n"
        " - run: manage.py migrate_blog_categories --from=categories.Category --to=categories_i18n.Category\n"
        " - remove 'categories' from INSTALLED_APPS.\n"
    )

    def create_parser(self, *args, **kwargs):
        parser = super(Command, self).create_parser(*args, **kwargs)
        parser.formatter_class = RawTextHelpFormatter
        return parser

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        parser.add_argument('-f', '--from', action='store', dest='from', help="The old model to read data from")
        parser.add_argument('-t', '--to', action='store', dest='to', help="The new model to migrate to")

    def handle(self, *args, **options):
        if args:
            raise CommandError("Command doesn't accept any arguments")

        Entry = get_entry_model()
        CategoryM2M = Entry.categories.through
        old_fk = CategoryM2M._meta.get_field('category')
        CurrentModel = old_fk.remote_field.model
        self.stdout.write("Current Entry.categories model: <{0}.{1}>".format(
            CurrentModel._meta.app_label, CurrentModel._meta.object_name
        ))

        old = options['from']
        new = options['to']
        if not old or not new:
            raise CommandError("Expected --from and --to options")

        if old.lower() == 'categories.category' and 'categories' not in settings.INSTALLED_APPS:
            # Can't import it in a Django 1.8+ project.
            OldModel = DummyCategoryBase
        else:
            try:
                OldModel = apps.get_model(old)
            except LookupError as e:
                raise CommandError("Invalid --from value: {0}".format(e))

        if not issubclass(OldModel, MPTTModel):
            raise CommandError("Expected MPTT model for --from value")

        try:
            NewModel = apps.get_model(new)
        except LookupError as e:
            raise CommandError("Invalid --to value: {0}".format(e))

        if not issubclass(NewModel, MPTTModel):
            raise CommandError("Expected MPTT model for --to value")

        if NewModel.objects.all().exists():
            raise CommandError("New model already has records, it should be an empty table!")

        old_i18n = issubclass(OldModel, TranslatableModel)
        new_i18n = issubclass(NewModel, TranslatableModel)
        old_title = _detect_title_field(OldModel)
        new_title = _detect_title_field(NewModel)
        mptt_fields = "lft, rght, tree_id, level, parent_id"

        with transaction.atomic():
            if not old_i18n and not new_i18n:
                # Untranslated to untranslated
                self.stdout.write("* Copying category fields...")
                with connection.cursor() as cursor:
                    cursor.execute(
                        'INSERT INTO {new_model}(id, slug, {new_title}, {mptt_fields}})'
                        ' SELECT id, slug, {old_title}, {mptt_fields} FROM {old_model}'.format(
                            new_model=NewModel._meta.db_table,
                            new_title=new_title,
                            old_model=OldModel._meta.db_table,
                            old_title=old_title,
                            mptt_fields=mptt_fields,
                        ))
            elif not old_i18n and new_i18n:
                # Untranslated to translated
                # - base table fields
                with connection.cursor() as cursor:
                    self.stdout.write("* Copying category base fields...")
                    cursor.execute(
                        'INSERT INTO {new_model}(id, {mptt_fields})'
                        ' SELECT id, {mptt_fields} FROM {old_model}'.format(
                            new_model=NewModel._meta.db_table,
                            old_model=OldModel._meta.db_table,
                            mptt_fields=mptt_fields,
                        ))
                    # - create translations on fallback language
                    self.stdout.write("* Creating category translations...")
                    cursor.execute(
                        'INSERT INTO {new_translations}(master_id, language_code, slug, {new_title})'
                        ' SELECT id, %s, slug, {old_title} FROM {old_model}'.format(
                            new_translations=NewModel._parler_meta.root_model._meta.db_table,
                            new_title=new_title,
                            old_model=OldModel._meta.db_table,
                            old_title=old_title,
                        ), [appsettings.FLUENT_BLOGS_DEFAULT_LANGUAGE_CODE])
            elif old_i18n and not new_i18n:
                # Reverse, translated to untranslated. Take fallback only
                # Convert all fields back to the single-language table.
                self.stdout.write("* Copying category fields and fallback language fields...")
                for old_category in OldModel.objects.all():
                    translations = old_category.translations.all()
                    try:
                        # Try default translation
                        old_translation = translations.get(language_code=appsettings.FLUENT_BLOGS_DEFAULT_LANGUAGE_CODE)
                    except ObjectDoesNotExist:
                        try:
                            # Try internal fallback
                            old_translation = translations.get(language_code__in=('en-us', 'en'))
                        except ObjectDoesNotExist:
                            # Hope there is a single translation
                            old_translation = translations.get()

                    fields = dict(
                        id=old_category.id,
                        lft=old_category.lft,
                        rght=old_category.rght,
                        tree_id=old_category.tree_id,
                        level=old_category.level,
                        # parler fields
                        _language_code=old_translation.language_code,
                        slug=old_category.slug
                    )
                    fields[new_title] = getattr(old_translation, old_title)
                    NewModel.objects.create(**fields)

            elif old_i18n and new_i18n:
                # Translated to translated
                # - base table
                with connection.cursor() as cursor:
                    self.stdout.write("* Copying category base fields...")
                    cursor.execute(
                        'INSERT INTO {new_model}(id, {mptt_fields})'
                        ' SELECT id, {mptt_fields} FROM {old_model}'.format(
                            new_model=NewModel._meta.db_table,
                            old_model=OldModel._meta.db_table,
                            mptt_fields=mptt_fields,
                        ))
                    # - all translations
                    self.stdout.write("* Copying category translations...")
                    cursor.execute(
                        'INSERT INTO {new_translations}(master_id, language_code, slug, {new_title})'
                        ' SELECT id, languag_code, slug, {old_title} FROM {old_translations}'.format(
                            new_translations=NewModel._parler_meta.root_model._meta.db_table,
                            new_title=new_title,
                            old_translations=OldModel._parler_meta.root_model._meta.db_table,
                            old_title=old_title,
                        ), [appsettings.FLUENT_BLOGS_DEFAULT_LANGUAGE_CODE])
            else:
                raise NotImplementedError()  # impossible combination

            self.stdout.write("* Switching M2M foreign key constraints...")
            __, __, __, kwargs = old_fk.deconstruct()
            kwargs['to'] = NewModel
            new_fk = models.ForeignKey(**kwargs)
            new_fk.set_attributes_from_name(old_fk.name)
            with connection.schema_editor() as schema_editor:
                schema_editor.alter_field(CategoryM2M, old_fk, new_fk)

        self.stdout.write("Done.\n")
        self.stdout.write("You may now remove the old category app from your project, INSTALLED_APPS and database.\n")


@python_2_unicode_compatible
class DummyCategoryBase(MPTTModel):
    """
    This base model includes the absolute bare bones fields and methods. One
    could simply subclass this model and do nothing else and it should work.
    """
    parent = TreeForeignKey('self', blank=True, null=True, related_name='children', on_delete=models.CASCADE)
    name = models.CharField(max_length=100)
    slug = models.SlugField()

    class Meta:
        db_table = 'categories_category'
        managed = False

    def __str__(self):
        return self.name


def _detect_title_field(Model):
    field_names = [f.name for f in Model._meta.get_fields()]

    if 'name' in field_names:
        return 'name'
    elif 'title' in field_names:
        return 'title'

    if issubclass(Model, TranslatableModel):
        field_names = Model._parler_meta.get_translated_fields()
        if 'name' in field_names:
            return 'name'
        elif 'title' in field_names:
            return 'title'

    raise CommandError("No 'name' or 'title' field found in model <{0}.{1}>".format(
        Model.__module__, Model.__name__
    ))



from fluent_pages.extensions import PageTypePlugin, page_type_pool

from fluent_blogs.pagetypes.blogpage.models import BlogPage

from .admin import BlogPageAdmin


@page_type_pool.register
class BlogPagePlugin(PageTypePlugin):
    model = BlogPage
    model_admin = BlogPageAdmin
    urls = 'fluent_blogs.urls'



from django.conf import settings
from django.contrib.admin import widgets
from parler.models import TranslatableModel

from fluent_blogs import appsettings
from fluent_blogs.admin.abstractbase import (
    AbstractEntryBaseAdmin, AbstractTranslatableEntryBaseAdmin,
    SeoEntryAdminMixin)
from fluent_blogs.models import get_entry_model

EntryModel = get_entry_model()
_model_fields = [f.name for f in EntryModel._meta.get_fields()]

if issubclass(EntryModel, TranslatableModel):
    _entry_admin_base = AbstractTranslatableEntryBaseAdmin
    # Some of the mixin fields could appear in the translated model instead of the regular shared model.
    # Add those fields here, so they will be added to the fieldsets by default as well.
    _model_fields += EntryModel._parler_meta.get_all_fields()
    _is_translated = True
else:
    _entry_admin_base = AbstractEntryBaseAdmin
    _is_translated = False


class EntryAdmin(SeoEntryAdminMixin, _entry_admin_base):
    """
    The Django admin class for the blog model.
    When another blog model is used, the ``fieldsets`` include it's additional fields.
    You can also use :class:`AbstractEntryBaseAdmin` with an optional ``SeoEntryAdminMixin``,
    to control the admin field layout completley.
    """
    # Redefine the fieldset, because it will be extended with auto-detected fields.
    FIELDSET_GENERAL = (None, {
        'fields': ('title', 'slug', 'status',),  # is filled with ('contents', 'categories', 'tags', 'enable_comments') below
    })

    if 'meta_keywords' in _model_fields:
        fieldsets = [
            FIELDSET_GENERAL,
            SeoEntryAdminMixin.FIELDSET_SEO,
            AbstractEntryBaseAdmin.FIELDSET_PUBLICATION,
        ]
    else:
        fieldsets = [
            FIELDSET_GENERAL,
            AbstractEntryBaseAdmin.FIELDSET_PUBLICATION,
        ]

    list_filter = ['status']  # reset, is rebuilt below.
    html_fields = []  # auto filled with excerpt_text
    formfield_overrides = {}
    formfield_overrides.update(SeoEntryAdminMixin.formfield_overrides)
    formfield_overrides.update({
        'intro': {
            'widget': widgets.AdminTextareaWidget(attrs={'rows': 4})
        },
    })

    def add_view(self, request, form_url='', extra_context=None):
        extra_context = extra_context or {}
        extra_context['html_fields'] = self.html_fields
        return super(EntryAdmin, self).add_view(request, form_url=form_url, extra_context=extra_context)

    def change_view(self, request, object_id, form_url='', extra_context=None):
        extra_context = extra_context or {}
        extra_context['html_fields'] = self.html_fields
        return super(EntryAdmin, self).change_view(request, object_id, form_url=form_url, extra_context=extra_context)


# Add all optional mixin fields
for _f in appsettings.FLUENT_BLOGS_ADMIN_FIELDS:
    if _f in _model_fields:
        EntryAdmin.FIELDSET_GENERAL[1]['fields'] += (_f,)

# Add filters for optional mixin fields
# Note, not adding 'tags' yet. It should only display tags that are in use, sorted by count.
if 'categories' in _model_fields:
    EntryAdmin.list_filter.append('categories')
if _is_translated and getattr(settings, 'PARLER_LANGUAGES', None):
    EntryAdmin.list_filter.append('translations__language_code')
if 'enable_comments' in _model_fields:
    EntryAdmin.list_filter.append('enable_comments')

# Autodetect HTML fields of known built-in mixins
if 'excerpt_text' in _model_fields:
    EntryAdmin.html_fields.append('excerpt_text')



from django.contrib import admin
from fluent_pages.integration.fluent_contents.admin import \
    FluentContentsPageAdmin

from .models import BlogPage


@admin.register(BlogPage)
class BlogPageAdmin(FluentContentsPageAdmin):
    """
    Customized admin appearance for the blogpage model.
    """
    placeholder_layout_template = "fluent_blogs/entry_archive_index.html"



from django.test import TestCase

from fluent_blogs.models import Entry, get_entry_model


class ModelTests(TestCase):

    def test_get_entry_model(self):
        self.assertIs(get_entry_model(), Entry)



from django.contrib import admin

from fluent_blogs.models import Entry, get_entry_model

from .abstractbase import (AbstractEntryBaseAdmin,
                           AbstractTranslatableEntryBaseAdmin,
                           SeoEntryAdminMixin)
from .entryadmin import EntryAdmin
from .forms import (AbstractEntryBaseAdminForm,
                    AbstractTranslatableEntryBaseAdminForm)

__all__ = (
    'AbstractEntryBaseAdminForm',
    'AbstractTranslatableEntryBaseAdminForm',

    'AbstractEntryBaseAdmin',
    'AbstractTranslatableEntryBaseAdmin',
    'SeoEntryAdminMixin',
    'EntryAdmin',
)


# Ony register the admin when the entry model is not customized.
if get_entry_model() is Entry:
    admin.site.register(Entry, EntryAdmin)



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from parler import appsettings as parler_appsettings
from parler.utils import is_supported_django_language, normalize_language_code

if 'categories' in settings.INSTALLED_APPS:
    # old django-categories style. Keep first for backwards compatibility
    _default_category_model = 'categories.Category'
elif 'categories_i18n' in settings.INSTALLED_APPS:
    _default_category_model = 'categories_i18n.Category'
else:
    # Default to new.
    _default_category_model = 'categories_i18n.Category'


# Configurable permalink style, yeah!
FLUENT_BLOGS_ENTRY_LINK_STYLE = getattr(settings, "FLUENT_BLOGS_ENTRY_LINK_STYLE", '/{year}/{month}/{slug}/')

# Configurable admin fields
# By default, all optional mixin fields are detected.
FLUENT_BLOGS_EXTRA_ADMIN_FIELDS = getattr(settings, 'FLUENT_BLOGS_EXTRA_ADMIN_FIELDS', ())
FLUENT_BLOGS_ADMIN_FIELDS = getattr(settings, 'FLUENT_BLOGS_ADMIN_FIELDS', tuple(FLUENT_BLOGS_EXTRA_ADMIN_FIELDS) + ('excerpt_image', 'excerpt_text', 'contents', 'categories', 'tags', 'enable_comments'))

# Advanced settings
FLUENT_BLOGS_FILTER_SITE_ID = getattr(settings, 'FLUENT_BLOGS_FILTER_SITE_ID', True)
FLUENT_BLOGS_BASE_TEMPLATE = getattr(settings, "FLUENT_BLOGS_BASE_TEMPLATE", 'fluent_blogs/base.html')
FLUENT_BLOGS_CATEGORY_MODEL = getattr(settings, "FLUENT_BLOGS_CATEGORY_MODEL", _default_category_model)
FLUENT_BLOGS_ENTRY_MODEL = getattr(settings, "FLUENT_BLOGS_ENTRY_MODEL", 'fluent_blogs.Entry')

# RSS feeds
FLUENT_BLOGS_FEED_PROTOCOL = getattr(settings, "FLUENT_BLOGS_FEED_PROTOCOL", 'http')
FLUENT_BLOGS_MAX_FEED_ITEMS = getattr(settings, "FLUENT_BLOGS_MAX_FEED_ITEMS", 30)

# Comment settings
FLUENT_BLOGS_INCLUDE_STATIC_FILES = getattr(settings, "FLUENT_BLOGS_INCLUDE_STATIC_FILES", True)
FLUENT_BLOGS_PAGINATE_BY = getattr(settings, "FLUENT_BLOGS_PAGINATE_BY", 10)

# Performance settings
FLUENT_BLOGS_PREFETCH_TRANSLATIONS = getattr(settings, 'FLUENT_BLOGS_PREFETCH_TRANSLATIONS', False)

# Note: the default language setting is used during the migrations
# Allow this module to have other settings, but default to the shared settings
FLUENT_DEFAULT_LANGUAGE_CODE = getattr(settings, 'FLUENT_DEFAULT_LANGUAGE_CODE', parler_appsettings.PARLER_DEFAULT_LANGUAGE_CODE)
FLUENT_BLOGS_DEFAULT_LANGUAGE_CODE = getattr(settings, 'FLUENT_BLOGS_DEFAULT_LANGUAGE_CODE', FLUENT_DEFAULT_LANGUAGE_CODE)
FLUENT_BLOGS_LANGUAGES = getattr(settings, 'FLUENT_BLOGS_LANGUAGES', parler_appsettings.PARLER_LANGUAGES)


# Clean settings
FLUENT_BLOGS_DEFAULT_LANGUAGE_CODE = normalize_language_code(FLUENT_BLOGS_DEFAULT_LANGUAGE_CODE)

if not is_supported_django_language(FLUENT_BLOGS_DEFAULT_LANGUAGE_CODE):
    raise ImproperlyConfigured("FLUENT_BLOGS_DEFAULT_LANGUAGE_CODE '{0}' does not exist in LANGUAGES".format(FLUENT_BLOGS_DEFAULT_LANGUAGE_CODE))

FLUENT_BLOGS_LANGUAGES = parler_appsettings.add_default_language_settings(
    FLUENT_BLOGS_LANGUAGES, 'FLUENT_BLOGS_LANGUAGES',
    code=FLUENT_BLOGS_DEFAULT_LANGUAGE_CODE,
    fallback=FLUENT_BLOGS_DEFAULT_LANGUAGE_CODE,
    hide_untranslated=False,
)


def get_language_settings(language_code, site_id=None):
    """
    Return the language settings for the current site
    """
    if site_id is None:
        site_id = getattr(settings, 'SITE_ID', None)

    for lang_dict in FLUENT_BLOGS_LANGUAGES.get(site_id, ()):
        if lang_dict['code'] == language_code:
            return lang_dict

    return FLUENT_BLOGS_LANGUAGES['default']



import sys


if sys.version_info[0] >= 3:
    string_types = str,
else:
    string_types = basestring,


def python_2_unicode_compatible(klass):
    """
    A decorator that defines __unicode__ and __str__ methods under Python 2.
    Under Python 3 it does nothing.

    To support Python 2 and 3 with a single code base, define a __str__ method
    returning text and apply this decorator to the class.
    """
    if sys.version_info[0] == 2:
        klass.__unicode__ = klass.__str__
        klass.__str__ = lambda self: self.__unicode__().encode('utf-8')
    return klass




from django.conf.urls import include, url
from django.contrib import admin

import fluent_blogs.urls

urlpatterns = [
    url(r'^admin/', admin.site.urls),
    url(r'^blog/', include(fluent_blogs.urls)),
]



from fluent_blogs.views.entries import (EntryArchiveIndex, EntryAuthorArchive,
                                        EntryCategoryArchive, EntryDayArchive,
                                        EntryDetail, EntryMonthArchive,
                                        EntryShortLink, EntryTagArchive,
                                        EntryYearArchive)

__all__ = (
    'EntryArchiveIndex', 'EntryYearArchive', 'EntryMonthArchive', 'EntryDayArchive',
    'EntryDetail', 'EntryShortLink', 'EntryCategoryArchive', 'EntryAuthorArchive', 'EntryTagArchive',
)



from ..base_models import (AbstractEntry, AbstractTranslatableEntry,
                           AbstractTranslatedFieldsEntry)
from ..managers import (EntryManager,  # noqa, old import paths
                        TranslatableEntryManager)
from .db import Entry, Entry_Translation, get_category_model, get_entry_model
from .query import get_category_for_slug

__all__ = (
    # Default translated models.
    'Entry',
    'Entry_Translation',

    # Default base models for classic and translated models.
    'AbstractEntry',
    'AbstractTranslatableEntry',
    'AbstractTranslatedFieldsEntry',

    # Utils for custom models.
    'get_entry_model',
    'get_category_model',
    'get_category_for_slug',
)
